# (c) 2018-2021 Andreas Motl <andreas.motl@panodata.org>
# License: GNU Affero General Public License, Version 3
"""
Introduction
============

This file hosts all built-in exposure sheets, i.e. multiple projects.
Its purpose is to define how to run different animation scenarios/sequences.

    An exposure sheet (also known variously as "dope sheet", "camera
    instruction sheet", or "X-sheet") is a sheet of paper used primarily in
    traditional animation to mark out the timing of various actions and
    dialogue.


Details
=======

The parameters `start`, `stop` and `every` should be self-explaining.

The parameter `mode` can have two values (defaulting to `window`):
- "window" will slide a window through the defined time range, with `every` as window width
- "cumulative" will use a fixed start time and stepwise expand the end time by `every`
"""

import logging
from datetime import datetime

from grafanimate.model import AnimationScenario, AnimationSequence, SequencingMode

logger = logging.getLogger(__name__)


def playdemo():
    """
    Run demo on `play.grafana.org`.

    Synopsis::

        grafanimate --scenario=playdemo --output=./animations
    """
    logger.info("Running scenario playdemo")

    return AnimationScenario(
        grafana_url="https://play.grafana.org/",
        dashboard_uid="000000012",
        sequences=[
            AnimationSequence(
                start=datetime(2021, 11, 14, 2, 0, 0),
                # Produce video with reasonable duration to explore different rendering options.
                stop=datetime(2021, 11, 14, 4, 16, 36),
                # Produce very short video.
                # stop=datetime(2021, 11, 14, 2, 16, 36),
                every="5min",
                mode=SequencingMode.CUMULATIVE,
            ),
        ],
    )


def playdemo_advanced():
    """
    Run demo on `play.grafana.org`, demonstrating different flavors of timestamps.

    Synopsis::

        grafanimate --scenario=playdemo_advanced --output=./animations
    """
    logger.info("Running scenario playdemo")

    return AnimationScenario(
        grafana_url="https://play.grafana.org/",
        dashboard_uid="000000012",
        sequences=[
            AnimationSequence(
                start=datetime(2021, 11, 14, 2, 0, 0),
                stop=datetime(2021, 11, 14, 2, 16, 36),
                every="5min",
                mode=SequencingMode.CUMULATIVE,
            ),
            AnimationSequence(
                start="2021-11-15T02:12:05Z",
                stop="2021-11-15T02:37:36Z",
                every="3min",
                mode=SequencingMode.CUMULATIVE,
            ),
            AnimationSequence(
                start=1637091011,
                stop=1637091911,
                every="4m5s",
                mode=SequencingMode.CUMULATIVE,
            ),
            AnimationSequence(
                start="-30m",
                stop="+30m",
                every="5m",
                mode=SequencingMode.CUMULATIVE,
            ),
            AnimationSequence(
                start="-14d",
                stop="start+7d",
                every="1d",
                mode=SequencingMode.CUMULATIVE,
            ),
            AnimationSequence(
                start="-14d",
                stop="now",
                every="1d",
                mode=SequencingMode.CUMULATIVE,
            ),
        ],
    )


def ldi_all():
    """
    Luftdaten.info, all

    Synopsis::

        grafanimate --grafana-url=https://daq.example.org/grafana --dashboard-uid=1aOmc1sik --scenario=ldi_all --output=./animations
    """
    logger.info("Running scenario ldi_all")

    return [
        # LDI, ramp-up
        AnimationSequence(
            start=datetime(2015, 10, 1),
            stop=datetime(2017, 1, 1),
            every="monthly",
        ),
        # LDI, growth
        AnimationSequence(
            start=datetime(2017, 1, 1),
            stop=datetime.now(),
            every="weekly",
        ),
    ]


def ldi_with_gaps():
    """
    Luftdaten.info, growth

    Synopsis::

        grafanimate --grafana-url=https://daq.example.org/grafana --dashboard-uid=1aOmc1sik --scenario=ldi_with_gaps --output=./animations
    """
    logger.info("Running scenario ldi_with_gaps")

    return [
        # LDI, ramp-up
        AnimationSequence(
            start=datetime(2015, 10, 1),
            stop=datetime(2017, 1, 1),
            every="monthly",
        ),
        # LDI, growth, with gap at 2018-04-29 - 2018-12-20
        # TODO: Detect empty data from datasource through Grafana Sidecar and skip respective images.
        AnimationSequence(
            start=datetime(2017, 1, 1),
            stop=datetime(2018, 6, 5),
            every="weekly",
        ),
        # LDI, until now
        AnimationSequence(
            start=datetime(2018, 12, 20),
            stop=datetime.now(),
            every="weekly",
        ),
    ]


def ldi_nye_2017_2018():
    """
    LDI, New Year's Eve 2018

    Synopsis::

        grafanimate --grafana-url=https://daq.example.org/grafana --dashboard-uid=1aOmc1sik --scenario=ldi_nye_2017_2018 --output=./animations
    """
    logger.info("Running scenario ldi_nye_2017_2018")
    return AnimationSequence(
        start=datetime(2017, 12, 31, 21, 0, 0),
        stop=datetime(2018, 1, 1, 4, 0, 0),
        every="10min",
    )


def ldi_nye_2018_2019():
    """
    LDI, New Year's Eve 2018/2019

    Synopsis::

        grafanimate --grafana-url=https://daq.example.org/grafana --dashboard-uid=1aOmc1sik --scenario=ldi_nye_2018_2019 --output=./animations
    """
    logger.info("Running scenario ldi_nye_2018_2019")
    return [
        AnimationSequence(
            start=datetime(2018, 12, 31, 15, 0, 0),
            stop=datetime(2018, 12, 31, 20, 0, 0),
            every="30min",
        ),
        AnimationSequence(
            start=datetime(2018, 12, 31, 20, 0, 0),
            stop=datetime(2018, 12, 31, 23, 0, 0),
            every="10min",
        ),
        AnimationSequence(
            start=datetime(2018, 12, 31, 23, 0, 0),
            stop=datetime(2019, 1, 1, 1, 0, 0),
            every="5min",
        ),
        AnimationSequence(
            start=datetime(2019, 1, 1, 1, 0, 0),
            stop=datetime(2019, 1, 1, 4, 0, 0),
            every="10min",
        ),
        AnimationSequence(
            start=datetime(2019, 1, 1, 4, 0, 0),
            stop=datetime(2019, 1, 1, 9, 0, 0),
            every="30min",
        ),
    ]


def cdc_maps():
    """
    DWD CDC, temperatur-sonne-and-niederschlag-karten

    Synopsis::

        grafanimate --grafana-url=https://daq.example.org/grafana --dashboard-uid=DLOlE_Rmz --scenario=cdc_maps --output=./animations
    """
    logger.info("Running scenario cdc_maps")
    return AnimationSequence(
        start=datetime(2018, 3, 6, 5, 0, 0),
        stop=datetime(2018, 3, 10, 23, 59, 59),
        every="hourly",
    )

    # Short sequence, for debugging processes.
    # return AnimationSequence(start=datetime(2018, 3, 6, 5, 0, 0), stop=datetime(2018, 3, 6, 6, 59, 59), every='hourly')


def uba_ldi_dwd_maps():
    """
    Labor: Studio / UBA/LDI/DWD-Studio [dev!]

    Synopsis::

        grafanimate --grafana-url=https://daq.example.org/grafana --dashboard-uid=of6c9qlmk --scenario=uba_ldi_dwd_maps --output=./animations
    """
    logger.info("Running scenario uba_ldi_dwd_maps")
    return AnimationSequence(
        start=datetime(2018, 10, 6, 5, 0, 0),
        stop=datetime(2018, 10, 10, 23, 59, 59),
        every="hourly",
    )


def ir_sensor_svg_pixmap():
    """
    start: 2018-08-14 03:16:00
    stop: 2018-08-14 03:16:36

    Synopsis::

        grafanimate --grafana-url=https://daq.example.org/grafana --dashboard-uid=acUXbj_mz --scenario=ir_sensor_svg_pixmap --output=./animations
    """
    logger.info("Running scenario ir_sensor_svg_pixmap")
    return AnimationSequence(
        start=datetime(2018, 8, 14, 3, 16, 0),
        stop=datetime(2018, 8, 14, 3, 16, 36),
        every="secondly",
    )



# (c) 2019 Andreas Motl <andreas@hiveeyes.org>
# License: GNU Affero General Public License, Version 3
from datetime import datetime, timedelta, timezone

from datetime_interval import Interval

"""
**********************************************
Improving time range control for "grafanimate"
**********************************************

Introduction
============
- https://community.hiveeyes.org/t/improving-time-range-control-for-grafanimate/1783/13
- https://en.wikipedia.org/wiki/Dope_sheet
- https://en.wikipedia.org/wiki/Exposure_sheet
- https://en.wikipedia.org/wiki/Animation#Stop_motion_animation

Running
=======
::

    python grafanimate/timecontrol.py

Backlog
=======
- Use datetime parser::

    from dateutil import parser

- Investigate NumPy
    - https://docs.scipy.org/doc/numpy/reference/arrays.datetime.html
    - https://jakevdp.github.io/PythonDataScienceHandbook/03.11-working-with-time-series.html

- Investigate Pandas
    - http://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Period.html
    - http://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.PeriodIndex.html
    - https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html
    - https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#time-span-representation
    - https://pandas.pydata.org/pandas-docs/stable/user_guide/timedeltas.html
    - http://www.marcelscharth.com/python/time.html

    - That's nice:
        - https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#custom-frequency-ranges
        - http://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.period_range.html#pandas.period_range
"""


class SlidingPeriodicInterval:
    def __init__(self, start, stop, every):
        self.start = start
        self.stop = stop
        self.every = every

    def __iter__(self):
        cursor = Interval(start=self.start, duration=self.every)
        yield cursor

        while cursor.end < self.stop:
            cursor += self.every
            yield cursor


class CumulativePeriodicInterval:
    def __init__(self, start, stop, every):
        self.start = start
        self.stop = stop
        self.every = every

    def __iter__(self):
        cursor = Interval(start=self.start, duration=self.every)
        yield cursor

        while cursor.end < self.stop:
            cursor.end += self.every
            yield cursor


def print_intervals(intervals):
    for interval in intervals:
        print(f"{interval.start} - {interval.end}")  # noqa: T201


def print_header(title):
    print()  # noqa: T201
    print("#", title)  # noqa: T201


def create_dope_sheet_blueprint():
    now = datetime.now(tz=timezone.utc)
    yesterday = now - timedelta(days=1)
    tomorrow = now + timedelta(days=1)

    print_header("Sliding forward")
    intervals = SlidingPeriodicInterval(
        start=yesterday,
        stop=tomorrow,
        every=timedelta(days=1),
    )
    print_intervals(intervals)

    # Just reversing the list of intervals yields deterministic results as it is literally
    # just the opposite of sliding forward without any different computation involved.
    print_header("Sliding reverse")
    intervals = SlidingPeriodicInterval(
        start=yesterday,
        stop=tomorrow,
        every=timedelta(days=1),
    )
    print_intervals(reversed(list(intervals)))

    print_header("Cumulative I (unaligned)")
    intervals = CumulativePeriodicInterval(
        start=yesterday,
        stop=tomorrow,
        every=timedelta(days=1),
    )
    print_intervals(intervals)

    print_header("Cumulative II (aligned)")
    now_aligned_to_hour = now - timedelta(
        minutes=now.minute,
        seconds=now.second,
        microseconds=now.microsecond,
    )
    intervals = CumulativePeriodicInterval(
        start=now_aligned_to_hour,
        stop=now_aligned_to_hour + timedelta(hours=2),
        every=timedelta(minutes=15),
    )
    print_intervals(intervals)


if __name__ == "__main__":
    create_dope_sheet_blueprint()



import os.path

from grafanimate.model import AnimationScenario, RenderingOptions
from grafanimate.postprocessing import MediaProducer
from grafanimate.util import ensure_directory, slug


def produce_artifacts(
    input,
    output,
    scenario: AnimationScenario,
    options: RenderingOptions,
):
    # TODO: Can use dashboard title as output filename here?
    # TODO: Can put `start` into filename?

    title_slug = slug(scenario.dashboard_title or scenario.dashboard_uid)
    scenario_slug = slug(scenario.source.replace(".py", ""))

    # Compute input pattern and output file name.
    input = os.path.join(str(input), "*.png")
    output = str(output).format(
        scenario=scenario_slug,
        title=title_slug,
        uid=scenario.dashboard_uid,
    )

    # Produce output artifacts.
    ensure_directory(output)
    producer = MediaProducer(options=options)
    return producer.render(source=input, target=output)



# (c) 2018-2021 Andreas Motl <andreas.motl@panodata.org>
# License: GNU Affero General Public License, Version 3
import logging
import os.path
import shutil
from tempfile import mkdtemp

from grafanimate.timeutil import format_date_filename

logger = logging.getLogger(__name__)


class TemporaryStorage:
    def __init__(self):
        self.workdir = mkdtemp()
        self.imagefile_template = "{uid}_{seq}_{start}_{stop}.png"

    def save_items(self, results) -> list[str]:
        files = []
        for item in results:
            # logger.info('item: %s', item)
            printable = item.copy()
            printable.data.image = printable.data.image[:23] + b"..."
            logger.debug("Item: %s", printable)
            files.append(self.save_item(item))
        return files

    def save_item(self, item) -> str:
        # Compute image sequence file name.
        imagename = self.imagefile_template.format(
            uid=item.meta.dashboard,
            seq=str(item.frame.sequence.index).zfill(4),
            start=format_date_filename(item.data.start),
            stop=format_date_filename(item.data.stop),
        )

        imagefile = os.path.join(self.workdir, imagename)

        # Store image.
        with open(imagefile, "wb") as f:
            f.write(item.data.image)

        logger.info(f"Saved frame to {imagefile} (size={len(item.data.image)})")

        return imagefile

    def __del__(self):
        shutil.rmtree(self.workdir)



# (c) 2018-2021 Andreas Motl <andreas.motl@panodata.org>
# License: GNU Affero General Public License, Version 3
import logging
import time

from munch import Munch, munchify

from grafanimate.grafana import GrafanaWrapper
from grafanimate.model import AnimationFrame, AnimationSequence

logger = logging.getLogger(__name__)


class SequentialAnimation:
    def __init__(
        self,
        grafana: GrafanaWrapper,
        dashboard_uid: str = None,
        options: Munch = None,
    ):
        self.grafana = grafana
        self.dashboard_uid = dashboard_uid
        self.options = options or None
        self.dry_run: bool = self.options.get("dry-run", False)

    def start(self):
        self.log("Opening dashboard")
        self.grafana.open_dashboard(self.dashboard_uid, options=self.options)
        self.log("Dashboard ready")

    def log(self, message):
        logger.info(message)
        self.grafana.console_info(message)

    def run(self, sequence: AnimationSequence):
        if not isinstance(sequence, AnimationSequence):
            return

        self.log(f"Starting animation: {sequence}")

        frame: AnimationFrame = None
        for frame in sequence.get_frames():
            # logger.info("=" * 42)

            # Render image.
            image = self.render(frame)

            # Build item model.
            item = munchify(
                {
                    "meta": {
                        "grafana": self.grafana,
                        "scenario": self.options["scenario"],
                        "dashboard": self.dashboard_uid,
                        "every": frame.timerange.recurrence.every,
                    },
                    "data": {
                        "start": frame.timerange.start,
                        "stop": frame.timerange.stop,
                        "image": image,
                    },
                    "frame": frame,
                },
            )

            yield item

        self.log("Animation finished")

    def render(self, frame: AnimationFrame):
        logger.debug("Adjusting time range control")
        self.grafana.timewarp(frame, self.dry_run)

        logger.debug("Rendering image")
        if self.options["exposure-time"] > 0:
            logger.info(
                "Waiting for %s seconds (exposure time)",
                self.options["exposure-time"],
            )
            time.sleep(self.options["exposure-time"])

        if self.options["panel-id"]:
            logger.debug("Updating Panel Label")
            self.grafana.update_tags()
        if not self.dry_run:
            return self.make_image()

    def make_image(self):
        image = self.grafana.render_image()
        # logger.info('Image size: %s', len(image))
        return image



# (c) 2018-2021 Andreas Motl <andreas.motl@panodata.org>
# License: GNU Affero General Public License, Version 3
import json
import logging
import time
from importlib.resources import read_text

from marionette_driver import Wait
from marionette_driver.errors import TimeoutException

from grafanimate.marionette import FirefoxMarionetteBase
from grafanimate.model import AnimationFrame
from grafanimate.timeutil import format_date_grafana

log = logging.getLogger(__name__)


class GrafanaWrapper(FirefoxMarionetteBase):
    """
    https://marionette-client.readthedocs.io/en/master/interactive.html
    """

    def __init__(
        self,
        baseurl: str = None,
        use_panel_events: bool = True,
        dry_run: bool = False,
    ):
        self.baseurl = baseurl
        self.use_panel_events = use_panel_events
        self.dry_run = dry_run
        log.info("Starting GrafanaWrapper on %s", baseurl)
        FirefoxMarionetteBase.__init__(self)

    def boot_grafana(self):
        """
        Navigate to Grafana application and inject Grafana Sidecar service.
        """
        log.info("Starting Grafana at %s", self.baseurl)
        self.set_window_size(1920, 1080)

        self.navigate(self.baseurl)

        rect = self.get_window_rect()
        self.marionette.set_window_rect(height=rect["height"], width=rect["width"])

    def navigate(self, url):
        # Navigate to resource URL.
        self.marionette.navigate(url)

        # Wait for Grafana application to load.
        self.wait_for_grafana()

        # Load Javascript for GrafanaStudio sidecar service.
        jsfiles = ["grafana-util.js", "grafana-studio.js"]
        for jsfile in jsfiles:
            javascript = read_text("grafanimate", jsfile)
            self.run_javascript(javascript)

    def wait_for_grafana(self):
        """
        Wait for element <grafana-app> to appear.
        """
        log.info("Waiting for Grafana to load")
        element = self.wait_for_element_tag("grafana-app")
        time.sleep(0.3)
        log.info("Grafana loaded")
        return element

    def login(self, username, password):
        log.info("Logging in")
        javascript = mkjscall("grafanaStudio.login", username, password)
        self.run_javascript(javascript)
        # the login works, but it is not shown even after reload
        # self.navigate(self.baseurl)

    def open_dashboard(self, uid, options=None):
        """
        Open a Grafana Dashboard with the designated uid
        and wait for all data to load into all panels.
        """
        options = options or {}
        javascript = mkjscall("grafanaStudio.openDashboard", uid, options)
        self.run_javascript(javascript)
        self.wait_for_frame_finished()

    def wait_for_frame_finished(self):
        if self.use_panel_events:
            return self.wait_all_data_received()

    def wait_all_data_received(self):
        """
        Wait for all data to arrive in the dashboard.
        """

        log.info('Waiting for "all-data-received" event')
        waiter = Wait(self.marionette, timeout=20.0, interval=0.1)

        def condition(marionette):  # noqa: ARG001
            return self.calljs("grafanaStudio.hasAllData", silent=True)

        try:
            waiter.until(condition)
        except TimeoutException as ex:
            log.warning("Timed out waiting for data: %s. Continuing anyway.", ex)

    def update_tags(self):
        return self.calljs("grafanaStudio.improvePanelChrome")

    def timewarp(self, frame: AnimationFrame, dry_run: bool = False):
        """
        Navigate the Dashboard to the designated point in time
        and wait for refreshing all child components including data.
        """

        # Notify user.
        message = f"Timewarp to {frame.timerange.start} -> {frame.timerange.stop}"
        log.info(message)
        self.console_log(message)

        # Perform timewarp.
        if not dry_run:
            self.timerange_set(
                format_date_grafana(frame.timerange.start, frame.timerange.recurrence),
                format_date_grafana(frame.timerange.stop, frame.timerange.recurrence),
            )
            self.wait_for_frame_finished()

    def timerange_set(self, starttime, endtime):
        """
        Adjust Grafana time control. This is not synchronous.
        """
        return self.calljs("grafanaStudio.setTime", starttime, endtime)

    def timerange_get(self):
        """
        For fetching the current timeRange values, use::

            var timeRange = angular.element('grafana-app').injector().get('timeSrv').timeRange();
            var temp_date_from = new Date(timeRange.from);
            var temp_date_to = new Date(timeRange.to);

        -- https://community.grafana.com/t/how-to-access-time-picker-from-to-within-a-text-panel-and-jquery/6071/3
        """
        return self.calljs("grafanaStudio.getTime")

    def run_javascript(self, sourcecode, silent=False):
        """
        Run the designated Javascript code directly in the scope
        of the application under test. Don't use any sandboxing.

        :code: Plain Javascript source code.
        """
        # https://github.com/mozilla/geckodriver/issues/1067#issuecomment-347180427
        # https://github.com/devtools-html/har-export-trigger/issues/27#issuecomment-424777524
        if not silent:
            log.debug("Running Javascript: %s", sourcecode)
        return self.marionette.execute_script(
            sourcecode,
            sandbox=None,
            new_sandbox=False,
        )

    def calljs(self, name, *args, silent=False):
        return self.run_javascript(mkjscall(name, *args), silent=silent)

    def get_dashboard_title(self):
        return self.calljs("grafanaStudio.getDashboardTitle")

    def console_log(self, message):
        """
        Write a message to the Browser console (console.log).
        """
        return self.calljs("console.log", message, silent=True)

    def console_info(self, message):
        """
        Write a message to the Browser console (console.info).
        """
        return self.calljs("console.info", message, silent=True)

    def wait_for_map(self):
        """
        Wait for map element <panel-plugin-grafana-worldmap-panel ...> to appear.
        """
        log.debug("Waiting for map element to appear in DOM")
        # TODO: Make naming/addressing more universal.
        element = self.wait_for_element_tag("panel-plugin-grafana-worldmap-panel")
        log.info("Finished waiting for map element")
        log.info(element)
        return element


def mkjscall(name, *arguments, **flags):
    flags.setdefault("add_return", True)
    tpl = "{return}{name}({arguments});"
    tplvars = {
        "return": flags["add_return"] and "return " or "",
        "name": name,
        "arguments": mkjsargs(*arguments),
    }
    return tpl.format(**tplvars)


def mkjsargs(*arguments):
    return json.dumps(arguments).strip("[]")



# (c) 2018-2021 Andreas Motl <andreas.motl@panodata.org>
# License: GNU Affero General Public License, Version 3
import json
import logging
import os
import typing as t
from pathlib import Path

from docopt import DocoptExit, docopt

from grafanimate import __appname__, __version__
from grafanimate.core import get_scenario, make_grafana, run_animation_scenario
from grafanimate.media import produce_artifacts
from grafanimate.model import RenderingOptions
from grafanimate.util import asbool, normalize_options, setup_logging

if t.TYPE_CHECKING:
    from grafanimate.spool import TemporaryStorage


log = logging.getLogger(__name__)


def run():
    """
    Usage:
      grafanimate [options] [--target=<target>]...
      grafanimate --version
      grafanimate (-h | --help)

    Options:

      --scenario=<scenario>         Which scenario to run. Built-in scenarios are defined within the
                                    `scenarios.py` file, however you can easily define scenarios in
                                    custom Python files.

                                    Scenarios can be referenced by arbitrary entrypoints, like:

                                    - `--scenario=grafanimate.scenarios:playdemo`    (module, symbol)
                                    - `--scenario=grafanimate/scenarios.py:playdemo` (file, symbol)
                                    - `--scenario=playdemo` (built-in)

      --output=<path>               Output path. Required.
                                    Optionally, use environment variable `GRAFANIMATE_OUTPUT` instead.

      --grafana-url=<url>           Base URL to Grafana.
                                    If your Grafana instance is protected, please specify credentials
                                    within the URL, e.g. https://user:pass@www.example.org/grafana.

      --dashboard-uid=<uid>         Grafana dashboard UID.

    Layout and scene options:
      --use-panel-events            Whether to enable using Grafana's panel events. [default: false]
                                    Caveat: Does not work for "d-solo" panels

      --panel-id=<id>               Render single panel only by navigating to "panelId=<id>&fullscreen".
      --dashboard-view=<mode>       Use Grafana's "d-solo" view for rendering single panels without header.
                                    "d-solo" is incompatible with the "use-panel-events" option

      --header-layout=<layout>      The header rendering subsystem offers different modes
                                    for amending the vanilla Grafana user interface.
                                    Multiple modes can be combined.
                                    [default: large-font]

                                    - no-chrome:            Set kiosk mode, remove sidemenu and more chrome
                                    - large-font:           Use larger font sizes for title and datetime
                                    - collapse-datetime:    Collapse datetime into title
                                    - studio:               Apply studio modifications. This options aggregates
                                                            "no-chrome", "large-font" and "collapse-datetime".
                                    - no-folder:            Don't include foldername in title

                                    - no-title:             Turn off title widget
                                    - no-datetime:          Turn off datetime widget

      --datetime-format=<format>    Datetime format to use with header layouts like "studio".
                                    Examples: YYYY-MM-DD HH:mm:ss, YYYY, HH:mm.

                                    There are also some format presets available here:
                                    - human-date:           on 2018-08-14
                                    - human-time:           at 03:16:05
                                    - human-datetime:       on 2018-08-14 at 03:16:05

                                    When left empty, the default is determined by the configured interval.
      --headless                    Use headless firefox if set, else render with frontend. [default: false]
                                    Setting this to true allows to render whole dashboards.
                                    Otherwise the screensize determines the maximum height.

    Capturing options:
      --exposure-time=<seconds>     How long to wait for each frame to complete rendering. [default: 0.5]
                                    Caveat: Is ignored when use-panel-events is set

    Rendering options:
      --video-framerate=<rate>      Framerate to apply when recording the video. This value will get propagated
                                    to FFmpeg's `-framerate` parameter. [default: 2]
      --video-fps=<fps>             Frames per second to apply when recording the video. This value will get
                                    propagated into FFmpeg's output encoding options. [default: 25]
      --gif-fps=<fps>               Frames per second to apply when recording the animated gif, propagated into
                                    FFmpeg's `-filter_complex` options. [default: 10]
      --gif-width=<pixel>           Width of the gif in pixels. [default: 480]

      --dry-run                     Enable dry-run mode
      --debug                       Enable debug logging
      -h --help                     Show this screen


    Examples for scenario mode. Script your animation in file `scenarios.py`. The output files
    will be saved at `./var/spool/{scenario}/{dashboard-uid}`.

      # Use freely accessible `play.grafana.org` for demo purposes.
      grafanimate --scenario=playdemo --output=./animations

      # Example for generating Luftdaten.info graph & map.
      export GRAFANIMATE_OUTPUT=./animations
      grafanimate --grafana-url=http://localhost:3000/ --dashboard-uid=1aOmc1sik --scenario=ldi_all

      # Use more parameters to control the rendering process.
      grafanimate --grafana-url=http://localhost:3000/ --dashboard-uid=acUXbj_mz --scenario=ir_sensor_svg_pixmap --header-layout=studio --datetime-format=human-time --panel-id=6 --use-panel-events

    """

    # Parse command line arguments.
    options = docopt(run.__doc__, version=__appname__ + " " + __version__)
    options = normalize_options(options, lists=["header-layout"])

    # Setup logging.
    debug = options.get("debug")
    log_level = logging.INFO
    if debug:
        log_level = logging.DEBUG
    setup_logging(log_level)

    # Debug command line options.
    if debug:
        log.info(f"Options: {json.dumps(options, indent=4)}")

    # Sanity checks.
    if not options["scenario"]:
        raise DocoptExit("Error: Parameter --scenario is mandatory")

    output_path = options.output
    if not output_path:
        output_path = os.environ.get("GRAFANIMATE_OUTPUT")
    if not output_path:
        raise DocoptExit(
            "Error: Parameter --output or environment variable GRAFANIMATE_OUTPUT is mandatory",
        )

    if options["dashboard-view"] == "d-solo" and not options["panel-id"]:
        raise DocoptExit(
            "Error: Parameter --panel-id is mandatory for --dashboard-view=d-solo",
        )

    options["exposure-time"] = float(options["exposure-time"])
    options["use-panel-events"] = asbool(options["use-panel-events"])
    options["headless"] = asbool(options["headless"])
    if options["use-panel-events"]:
        options["exposure-time"] = 0

    # Prepare rendering options.
    render_options = RenderingOptions(
        video_framerate=int(options.video_framerate),
        video_fps=int(options.video_fps),
        gif_fps=int(options.gif_fps),
        gif_width=int(options.gif_width),
    )

    # Load scene.
    scenario = get_scenario(options["scenario"])

    # Resolve URL to Grafana, either from command line (precedence), or from scenario file.
    if options["grafana-url"]:
        scenario.grafana_url = options["grafana-url"]
    if not scenario.grafana_url:
        scenario.grafana_url = "http://localhost:3000"

    # The dashboard UID can be defined either in the scenario or via command line.
    # Command line takes precedence.
    if options["dashboard-uid"]:
        scenario.dashboard_uid = options["dashboard-uid"]
    if not scenario.dashboard_uid:
        raise KeyError(
            "Dashboard UID is mandatory, either supply it on the command line or via scenario file",
        )

    # Open a Grafana site in Firefox, using Marionette.
    grafana = make_grafana(
        scenario.grafana_url,
        scenario.dashboard_uid,
        options,
        options["headless"],
    )

    # Invoke pipeline: Run stop motion animation, producing single frames.
    storage: TemporaryStorage = run_animation_scenario(
        scenario=scenario,
        grafana=grafana,
        options=options,
    )

    # Define output filename pattern.
    output = Path(output_path) / "{scenario}--{title}--{uid}.mp4"

    # Run rendering sequences, produce composite media artifacts.
    scenario.dashboard_title = grafana.get_dashboard_title()
    if not options.dry_run:
        results = produce_artifacts(
            input=storage.workdir,
            output=output,
            scenario=scenario,
            options=render_options,
        )
        log.info("Produced %s results\n%s", len(results), json.dumps(results, indent=2))



# (c) 2018-2021 Andreas Motl <andreas.motl@panodata.org>
# License: GNU Affero General Public License, Version 3
import importlib
import logging
from pathlib import Path

from furl import furl
from munch import Munch

from grafanimate.animations import SequentialAnimation
from grafanimate.grafana import GrafanaWrapper
from grafanimate.model import AnimationScenario, AnimationSequence
from grafanimate.spool import TemporaryStorage
from grafanimate.util import as_list, filter_dict, import_module

log = logging.getLogger(__name__)


def make_grafana(
    url: str,
    dashboard_uid: str,
    options: dict,
    headless=False,
) -> GrafanaWrapper:
    do_login = False
    url = furl(url)
    if url.username:
        do_login = True
        username = url.username
        password = url.password
        url.username = None
        url.password = None

    view = "d"
    slug = "foo"
    query = ""
    if options["dashboard-view"]:
        view = options["dashboard-view"]
    if options["panel-id"]:
        if options["dashboard-view"] == "d-solo":
            query = (
                "?panelId="
                + options["panel-id"]
                + "&__feature.dashboardSceneSolo&fullscreen"
            )
        else:
            query = "?viewPanel=" + options["panel-id"]

    if str(url)[-1] != "/":
        url = str(url) + "/"
    url = str(url) + view + "/" + dashboard_uid + "/" + slug + query
    log.info(f"URL: {url}")

    grafana = GrafanaWrapper(
        baseurl=str(url),
        use_panel_events=options["use-panel-events"],
    )
    grafana.boot_firefox(headless=headless)
    grafana.boot_grafana()

    if do_login:
        grafana.login(username, password)

    return grafana


def get_scenario(source: str) -> AnimationScenario:
    """
    Resolve scenario from Python module or file.
    """

    # If it's not a full-qualified reference, fall back to trying the built-in scenario methods.
    modname, _, symbol = source.partition(":")
    if not symbol:
        symbol = modname
        modname = "grafanimate.scenarios"

    # Load module and resolve symbol.
    module = load_module(modname)
    scenario: AnimationScenario = resolve_reference(module, symbol)

    if scenario is None:
        raise NotImplementedError(
            f'Animation scenario "{source}" not found or implemented',
        )

    scenario.source = source

    return scenario


def load_module(modname):
    if Path(modname).exists():
        module = import_module("<unknown>", modname)
    else:
        module = importlib.import_module(modname)
    return module


def resolve_reference(module, symbol):
    reference = getattr(module, symbol, None)
    if callable(reference):
        reference = reference()
    if isinstance(reference, AnimationScenario):
        pass
    elif isinstance(reference, (AnimationSequence, list)):
        reference = AnimationScenario(sequences=as_list(reference))
    return reference


def run_animation_scenario(
    scenario: AnimationScenario,
    grafana: GrafanaWrapper,
    options: Munch,
) -> TemporaryStorage:
    log.info(
        f"Running animation scenario at {scenario.grafana_url}, with dashboard UID {scenario.dashboard_uid}",
    )

    storage = TemporaryStorage()

    # Define options to be propagated to the Javascript client domain.
    animation_options = filter_dict(
        options,
        [
            "panel-id",
            "dashboard-view",
            "header-layout",
            "datetime-format",
            "exposure-time",
            "use-panel-events",
            "scenario",
        ],
    )

    # Start the engines.
    animation = SequentialAnimation(
        grafana=grafana,
        dashboard_uid=scenario.dashboard_uid,
        options=animation_options,
    )
    animation.start()

    # Run animation scenario.
    for index, sequence in enumerate(scenario.sequences):
        sequence.index = index
        results = list(animation.run(sequence))
        if not options.dry_run:
            storage.save_items(results)

    return storage


def run_animation_adhoc():
    # TODO: Introduce ad-hoc mode. In the meanwhile, please use scenario mode.
    """
    animator = SequentialAnimation(
        options['url'],
        time_start=options.get('start'),
        time_end=options.get('end', 'now'),
        time_step=options.get('every', '1h')
    )
    animator.run()
    """



# (c) 2018-2021 Andreas Motl <andreas@hiveeyes.org>
# License: GNU Affero General Public License, Version 3
import logging
import os

from grafanimate.model import RenderingOptions

logger = logging.getLogger(__name__)


class MediaProducer:
    def __init__(self, options: RenderingOptions):
        self.options = options

    def to_video(self, source, target):
        """
        http://hamelot.io/visualization/using-ffmpeg-to-convert-a-set-of-images-into-a-video/
        https://stackoverflow.com/questions/24961127/how-to-create-a-video-from-images-with-ffmpeg

        https://stackoverflow.com/questions/38726370/ffmpeg-text-watermark-bottom-left/38728172#38728172
        https://superuser.com/questions/939357/ffmpeg-watermark-on-bottom-right-corner/939386#939386
        """
        # command = "ffmpeg -framerate 4 -pattern_type glob -i '{}' -c:v libx264 -r 30 -pix_fmt yuv420p '{}' -y".format(source, target)
        # -vf "fps=25,format=yuv420p,drawtext=fontfile=OpenSans-Regular.ttf:text='Title of this Video':fontcolor=white:fontsize=24:x=(w-tw)/2:y=(h/PHI)+th
        # command = "ffmpeg -framerate 4 -pattern_type glob -i '{}' -c:v libx264 -vf 'fps=25,format=yuv420p,drawtext=text=Produced with Grafana and grafanimate:fontsize=11:x=w-tw-30:y=h-th-10:fontcolor=lightgrey:fontfile=/Library/Fonts/Arial.ttf' '{}' -y".format(source, target)

        # TODO: Expose `-framerate` and `fps` values.
        # use the `pad` option to avoid ffmpeg errors like 'height not divisible by 2'
        command = f"ffmpeg -framerate {self.options.video_framerate} -pattern_type glob -i '{source}' -c:v libx264 -vf 'pad=ceil(iw/2)*2:ceil(ih/2)*2,fps={self.options.video_fps},format=yuv420p' '{target}' -y"
        logger.info(f"Rendering video: {target}")
        logger.debug(command)
        os.system(command)  # noqa: S605

    def to_gif(self, source, target):
        """
        # High Quality Gifs with FFmpeg
        # https://medium.com/@colten_jackson/doing-the-gif-thing-on-debian-82b9760a8483
        ::

            ffmpeg -ss 2.6 -t 1.3 -i MVI_7035.MOV \
                -vf fps=15,scale=320:-1:flags=lanczos,palettegen palette.png

            ffmpeg -ss 2.6 -t 1.3 -i MVI_7035.MOV -i palette.png \
                -filter_complex "fps=15,scale=400:-1:flags=lanczos[x];[x][1:v]paletteuse" sixthtry.gif

        > I think the -1 in the video filters refers to the height and basically means ‘preserve aspect ratio’.
        > So, in this case, 320:-1 means scale to: w=320, h=320/CurrentWidth*CurrentHeight

        # How to make GIFs with FFmpeg
        https://engineering.giphy.com/how-to-make-gifs-with-ffmpeg/

        # How do I convert a video to GIF using FFmpeg, with reasonable quality?
        https://superuser.com/questions/556029/how-do-i-convert-a-video-to-gif-using-ffmpeg-with-reasonable-quality

        # High quality GIF with FFmpeg
        http://blog.pkh.me/p/21-high-quality-gif-with-ffmpeg.html

        # FFmpeg gif script for bash
        https://github.com/thevangelist/FFMPEG-gif-script-for-bash

        > Here is a better version with fifo filter to avoid Buffer queue overflow when using paletteuse filter.
        > By using split filter to avoid the creation of intermediate palette PNG file.
        > -- https://superuser.com/questions/556029/how-do-i-convert-a-video-to-gif-using-ffmpeg-with-reasonable-quality/1256459#1256459

        ffmpeg -i input.mp4 -filter_complex 'fps=10,scale=320:-1:flags=lanczos,split [o1] [o2];[o1] palettegen [p]; [o2] fifo [o3];[o3] [p] paletteuse' out.gif
        """

        """
        # PoC
        ffmpeg -i dwd-cdc-2018-08.mov -vf fps=15,scale=320:-1:flags=lanczos,palettegen palette.png
        ffmpeg -i dwd-cdc-2018-08.mov -i palette.png -filter_complex "fps=15,scale=480:-1:flags=lanczos[x];[x][1:v]paletteuse" dwd-cdc-2018-08.gif -y
        """

        """
        # One-step conversion.
        ffmpeg -ss $START -i $IN_URL -t $LENGTH -filter_complex "fps=$FPS,scale=$WIDTH:-1:flags=lanczos,split=2 [a][b]; [a] palettegen [pal]; [b] [pal] paletteuse" out.gif
        ffmpeg -i dwd-cdc-2018-08.mov -filter_complex 'fps=10,scale=480:-1:flags=lanczos,split [o1] [o2];[o1] palettegen [p]; [o2] fifo [o3];[o3] [p] paletteuse' dwd-cdc-2018-08-v2.gif -y
        """

        # TODO: Expose `fps` and `scale` values.
        command = f"ffmpeg -i '{source}' -filter_complex 'fps={self.options.gif_fps},scale={self.options.gif_width}:-1:flags=lanczos,split [o1] [o2];[o1] palettegen [p]; [o2] [p] paletteuse' '{target}' -y"
        logger.info("Rendering GIF: %s", target)
        logger.debug(command)
        os.system(command)  # noqa: S605

    def upload_server(self, source):
        command = f"make --makefile=/Users/amo/dev/hiveeyes/sources/documentation/Makefile ptrace source={source}"
        os.system(command)  # noqa: S605

    def render(self, source, target):
        mp4 = target
        suffix = "." + target.split(".")[-1]
        gif = mp4.replace(suffix, ".gif")
        self.to_video(source, mp4)
        self.to_gif(mp4, gif)
        results = [mp4, gif]
        return results


def run(source, target):
    renderer = MediaProducer(options=RenderingOptions())
    renderer.render(source, target)
    # upload_server(gif)
    # upload_server(mp4)


if __name__ == "__main__":
    run("./var/spool/*_1aOmc1sik_*.png", "./var/results/ldi-coverage.mp4")

    # to_video('./var/spool/*_1aOmc1sik_*.png', 'ldi-coverage.mp4')
    # to_gif('ldi-coverage.mp4', 'ldi-coverage.gif')
    # upload_server('ldi-coverage.gif')
    # upload_server('ldi-coverage.mp4')



# (c) 2018-2021 Andreas Motl <andreas.motl@panodata.org>
# License: GNU Affero General Public License, Version 3
import logging
import os
import re
import socket
import sys
from contextlib import closing
from pathlib import Path

from munch import munchify
from unidecode import unidecode


def setup_logging(level=logging.INFO):
    log_format = "%(asctime)-10s [%(name)-30s] %(levelname)-7s: %(message)s"
    logging.basicConfig(format=log_format, stream=sys.stderr, level=level)

    # TODO: Control debug logging of HTTP requests through yet another commandline option "--debug-http" or "--debug-requests"
    requests_log = logging.getLogger("requests")
    requests_log.setLevel(logging.WARN)


def normalize_options(options, lists=None):
    lists = lists or []
    normalized = {}
    for key_raw, value in options.items():
        key = key_raw.strip("--<>")
        normalized[key] = normalized[key.replace("-", "_")] = value
    for key in lists:
        normalized[key] = normalized[key.replace("-", "_")] = read_list(normalized[key])
    return munchify(normalized)


def read_list(data, separator=","):
    if data is None:
        return []
    result = list(map(lambda x: x.strip(), data.split(separator)))
    if len(result) == 1 and not result[0]:
        result = []
    return result


def find_program_candidate(candidates):
    for candidate in candidates:
        if os.path.isfile(candidate):
            return candidate


def check_socket(host, port):
    # https://stackoverflow.com/questions/19196105/python-how-to-check-if-a-network-port-is-open-on-linux
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as sock:
        sock.settimeout(1)
        if sock.connect_ex((host, port)) == 0:
            return True
        return False


def filter_dict(data, keys):
    return {k: v for k, v in data.items() if k in keys}


def slug(text):
    """Make a URL-safe, human-readable version of the given text

    This will do the following:

    1. decode unicode characters into ASCII
    2. shift everything to lowercase
    3. strip whitespace
    4. replace other non-word characters with dashes
    5. strip extra dashes

    This somewhat duplicates the :func:`Google.slugify` function but
    slugify is not as generic as this one, which can be reused
    elsewhere.

    Stolen from beetsplug.lyrics.
    """
    return re.sub(r"\W+", "-", unidecode(text).lower().strip()).strip("-")


def ensure_directory(path):
    directory = os.path.dirname(path)
    if not os.path.exists(directory):
        os.makedirs(directory)


def asbool(obj):
    # from sqlalchemy.util.asbool
    if isinstance(obj, str):
        obj = obj.strip().lower()
        if obj in ["true", "yes", "on", "y", "t", "1"]:
            return True
        if obj in ["false", "no", "off", "n", "f", "0"]:
            return False
        raise ValueError(f"String is not true/false: {obj!r}")
    return bool(obj)


def is_sequence(seq):
    # From `numpy.distutils.misc_util`
    if isinstance(seq, str):
        return False
    try:
        len(seq)
    except Exception:
        return False
    return True


def as_list(seq):
    # From `numpy.distutils.misc_util`
    if is_sequence(seq):
        return list(seq)
    return [seq]


def import_module(name: str, path: str):
    """
    Import Python module from file.
    """

    # Use absolute path.
    modulefile = Path(path).absolute()

    # Import module.
    # https://stackoverflow.com/a/67692
    import importlib.util

    spec = importlib.util.spec_from_file_location(name, modulefile)
    if spec is None:
        raise FileNotFoundError(f"Unable to find module file {modulefile}")
    mod = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(mod)

    return mod



# (c) 2018-2021 Andreas Motl <andreas.motl@panodata.org>
# License: GNU Affero General Public License, Version 3
import dataclasses
from datetime import datetime, timedelta, timezone
from typing import Optional, Union

import dateutil.parser
from dateutil.relativedelta import relativedelta
from dateutil.rrule import DAILY, HOURLY, MINUTELY, MONTHLY, SECONDLY, WEEKLY, YEARLY
from pytimeparse2 import parse as parse_human_time


@dataclasses.dataclass
class RecurrenceInfo:
    """
    For feeding data to `dateutil.rrule.rrule`.
    """

    # One of `rrules`s DAILY, HOURLY, ...
    frequency: int

    # The segment duration in seconds.
    interval: int

    # The segment duration, expressed as `relativedelta`.
    duration: relativedelta

    # Original interval/windowing label.
    every: str = None


@dataclasses.dataclass
class Timerange:
    start: datetime
    stop: datetime
    recurrence: RecurrenceInfo


def get_freq_delta(every: str) -> RecurrenceInfo:
    rr_freq = MINUTELY
    rr_interval = 1

    # 1. Attempt to parse time using `pytimeparse` module.
    # https://pypi.org/project/pytimeparse/
    duration = parse_human_time(every)
    if duration:
        delta = get_relativedelta(seconds=duration)

        if delta.years:
            rr_freq = YEARLY
            rr_interval = delta.years
        elif delta.months:
            rr_freq = MONTHLY
            rr_interval = delta.months
        elif delta.days:
            rr_freq = DAILY
            rr_interval = delta.days
        elif delta.hours:
            rr_freq = HOURLY
            rr_interval = delta.hours
        elif delta.minutes:
            rr_freq = MINUTELY
            rr_interval = delta.minutes
        else:
            rr_freq = SECONDLY
            rr_interval = delta.seconds

        if rr_freq != SECONDLY:
            delta -= relativedelta(seconds=1)

        return RecurrenceInfo(
            every=every,
            frequency=rr_freq,
            interval=rr_interval,
            duration=delta,
        )

    # 2. Compute parameters from specific labels, expression periods.

    # Secondly
    if every == "secondly":
        rr_freq = SECONDLY
        delta = timedelta(seconds=1)

    # Minutely
    elif every == "minutely":
        rr_freq = MINUTELY
        delta = timedelta(minutes=1) - timedelta(seconds=1)

    # Each 5 minutes
    elif every == "5min":
        rr_freq = MINUTELY
        rr_interval = 5
        delta = timedelta(minutes=5) - timedelta(seconds=1)

    # Each 10 minutes
    elif every == "10min":
        rr_freq = MINUTELY
        rr_interval = 10
        delta = timedelta(minutes=10) - timedelta(seconds=1)

    # Each 30 minutes
    elif every == "30min":
        rr_freq = MINUTELY
        rr_interval = 30
        delta = timedelta(minutes=30) - timedelta(seconds=1)

    # Hourly
    elif every == "hourly":
        rr_freq = HOURLY
        delta = timedelta(hours=1) - timedelta(seconds=1)

    # Daily
    elif every == "daily":
        rr_freq = DAILY
        delta = timedelta(days=1) - timedelta(seconds=1)

    # Weekly
    elif every == "weekly":
        rr_freq = WEEKLY
        delta = timedelta(weeks=1) - timedelta(seconds=1)

    # Monthly
    elif every == "monthly":
        rr_freq = MONTHLY
        delta = relativedelta(months=+1) - relativedelta(seconds=1)

    # Yearly
    elif every == "yearly":
        rr_freq = YEARLY
        delta = relativedelta(years=+1) - relativedelta(seconds=1)

    else:
        raise ValueError(f'Unknown interval "{every}"')

    if isinstance(delta, timedelta):
        delta = get_relativedelta(seconds=delta.total_seconds())

    return RecurrenceInfo(
        every=every,
        frequency=rr_freq,
        interval=rr_interval,
        duration=delta,
    )


def get_relativedelta(seconds: int):
    # TODO: Add to `pytimeparse2`?
    # https://stackoverflow.com/questions/16977768/elegant-way-to-convert-python-datetime-timedelta-to-dateutil-relativedelta

    seconds_in = {
        "year": 365 * 24 * 60 * 60,
        "month": 30 * 24 * 60 * 60,
        "day": 24 * 60 * 60,
        "hour": 60 * 60,
        "minute": 60,
    }

    years, rem = divmod(seconds, seconds_in["year"])
    months, rem = divmod(rem, seconds_in["month"])
    days, rem = divmod(rem, seconds_in["day"])
    hours, rem = divmod(rem, seconds_in["hour"])
    minutes, rem = divmod(rem, seconds_in["minute"])
    seconds = rem

    return relativedelta(
        years=years,
        months=months,
        days=days,
        hours=hours,
        minutes=minutes,
        seconds=seconds,
    ).normalized()


def format_date_filename(date, every=None):  # noqa: ARG001
    # pattern = '%Y-%m-%d'
    pattern = "%Y-%m-%dT%H-%M-%S"
    # if every in ['secondly', 'minutely', 'hourly']:
    #    pattern = '%Y-%m-%dT%H-%M-%S'
    date_formatted = date.strftime(pattern)
    return date_formatted


def format_date_grafana(date: datetime, recurrence: RecurrenceInfo):
    pattern = "%Y-%m-%d"
    if recurrence.frequency in [SECONDLY, MINUTELY, HOURLY]:
        pattern = "%Y-%m-%dT%H:%M:%S"
    date_formatted = date.strftime(pattern)
    return date_formatted


def convert_absolute_timestamp(value: Union[datetime, str]) -> datetime:
    """
    Read and convert absolute timestamps.
    """
    if isinstance(value, datetime):
        pass
    elif isinstance(value, int):
        value = datetime.fromtimestamp(value, tz=timezone.utc)
    elif isinstance(value, str):
        value = dateutil.parser.parse(value)
    else:
        raise TypeError(
            f"Unknown data type for `start` or `stop` value: {value} ({type(value)})",
        )
    return value


def convert_input_timestamp(
    value: Union[datetime, str],
    relative_to: Optional[datetime] = None,
) -> datetime:
    """
    Read and convert absolute or relative (humanized) timestamps.
    """
    if isinstance(value, str):
        if value == "now":
            return datetime.now(tz=timezone.utc)
        try:
            delta = parse_human_time(value)
            if not delta:
                raise ValueError(f"Unable to parse {value}")
            return relative_to + timedelta(seconds=delta)
        except ValueError as ex:
            if "Unable to parse" not in str(ex):
                raise

    return convert_absolute_timestamp(value)



# (c) 2018-2021 Andreas Motl <andreas.motl@panodata.org>
# License: GNU Affero General Public License, Version 3
import dataclasses
import logging
from collections.abc import Generator
from datetime import datetime, timezone
from enum import Enum
from typing import Optional, Union

from dateutil.rrule import rrule

from grafanimate.timeutil import (
    RecurrenceInfo,
    Timerange,
    convert_input_timestamp,
    get_freq_delta,
)

logger = logging.getLogger(__name__)


class SequencingMode(Enum):
    WINDOW = "window"
    CUMULATIVE = "cumulative"


@dataclasses.dataclass
class AnimationFrame:
    sequence: "AnimationSequence"
    timerange: Timerange


class AnimationSequence:
    def __init__(
        self,
        start: Union[datetime, int, str],
        stop: Union[datetime, int, str],
        every: str = None,
        recurrence: RecurrenceInfo = None,
        mode: Optional[SequencingMode] = SequencingMode.WINDOW,
    ):
        # Convert start/stop timestamps, resolving relative timestamps.
        now = datetime.now(tz=timezone.utc)
        self.start = convert_input_timestamp(start, relative_to=now)
        if isinstance(stop, str) and stop.startswith("start"):
            _stop = stop.replace("start", "")
            self.stop = convert_input_timestamp(_stop, relative_to=self.start)
        else:
            self.stop = convert_input_timestamp(stop, relative_to=now)

        # Analyze `every` parameter and converge into `RecurrenceInfo`.
        # From `every` (interval designator), compute frequency, interval and delta.
        if recurrence is not None:
            self.recurrence = recurrence
        else:
            self.recurrence = get_freq_delta(every)

        self.mode = mode
        self.index = None

        # Sanity checks.
        if self.start > self.stop:
            message = f"Timestamp start={self.start.isoformat()} is after stop={self.stop.isoformat()}"
            raise ValueError(message)

    def get_frames(self) -> Generator[AnimationFrame, None, None]:
        timerange = Timerange(
            start=self.start,
            stop=self.stop,
            recurrence=self.recurrence,
        )

        # until = datetime.now()
        if self.mode == SequencingMode.CUMULATIVE:
            timerange.stop += self.recurrence.duration

        # Compute complete date range.
        logger.info(
            "Creating rrule: dtstart=%s, until=%s, freq=%s, interval=%s",
            timerange.start,
            timerange.stop,
            self.recurrence.frequency,
            self.recurrence.interval,
        )
        daterange = list(
            rrule(
                dtstart=timerange.start,
                until=timerange.stop,
                freq=self.recurrence.frequency,
                interval=self.recurrence.interval,
            ),
        )
        # logger.info('Date range is: %s', daterange)

        # Iterate date range.
        for date in daterange:
            # Compute start and end dates based on mode.

            if self.mode == SequencingMode.WINDOW:
                start = date
                stop = date + self.recurrence.duration

            elif self.mode == SequencingMode.CUMULATIVE:
                start = timerange.start
                stop = date

            frame = AnimationFrame(
                sequence=self,
                timerange=Timerange(start=start, stop=stop, recurrence=self.recurrence),
            )
            yield frame

    def get_timeranges_isoformat(self) -> Generator[str, None, None]:
        for frame in self.get_frames():
            item = f"{frame.timerange.start.isoformat()}/{frame.timerange.stop.isoformat()}"
            # print(f'"{item}",')
            yield item


@dataclasses.dataclass
class AnimationScenario:
    sequences: list[AnimationSequence]
    grafana_url: Optional[str] = None
    dashboard_uid: Optional[str] = None
    dashboard_title: Optional[str] = None
    source: Optional[str] = None


@dataclasses.dataclass
class RenderingOptions:
    video_framerate: int = 2
    video_fps: int = 25
    gif_fps: int = 10
    gif_width: int = 480
